/******************************************************************************
*
* project name: Game of Life Demo
* file name:    life.c 
* initial date: 28/08/2000
* author:       thomas.nussbaumer@gmx.net
* description:  turns your calculator screen content into a Game of Life
*               simulation
*
* compile with: tigcc -O2 life.c
*
* $Id: life.c,v 1.1 2000/08/28 17:50:34 Thomas Nussbaumer Exp $
*
******************************************************************************/

//-----------------------------------------------------------------------------
// NOTE: To keep things simple this program allocates two buffers which holds
//       one pixel per byte. This way we haven't to deal with bit extraction.
//       The penality for this is a 60kB memory usage.
//       
//       This program is neither efficient nor optimized for speed. Its just
//       a 30 minutes "what-should-I-do-with-my-lunch-break" hack.
//
// NOTE2: The handling of the 4 corner points is left out for simplicity !!!!
//
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// The Game of Life was invented by John Conway (as you might have gathered). 
// The game is played on a field of cells, each of which has eight neighbors 
// (adjacent cells). A cell is either occupied (by an organism) or not. 
//
// The rules for deriving a generation from the previous one are these: 
// ====================================================================
//
// [Death]    If an occupied cell has 0, 1, 4, 5, 6, 7, or 8 occupied neighbors, 
//            the organism dies
//            (0, 1 neighbors: of loneliness; 4 thru 8: of overcrowding). 
//
// [Survival] If an occupied cell has two or three neighbors, the organism 
//            survives to the next generation. 
//
// [Birth]    If an unoccupied cell has three occupied neighbors, it becomes 
//            occupied.  
//-----------------------------------------------------------------------------

#include <nostub.h>
#include <all.h>

int _ti92plus;
int _ti89;

unsigned char* source = 0;
unsigned char* dest   = 0;
int maxcol            = 0;
int maxline           = 0;

#define SVAL(l,c)  *(source+240*(l)+(c))
#define DVAL(l,c)  *(dest+240*(l)+(c))
    

//=============================================================================
// does one iteration 
//=============================================================================
void Process() {
   unsigned int i;
   unsigned char* in  = source;
   unsigned char* out = dest;
   
   unsigned int line,col;
   
   //-------------------------------------------------
   // handle inner content (border will handled later)
   //-------------------------------------------------
   for (line=1;line<maxline;line++) {
       for (col=1;col<maxcol;col++) {
           int sum = SVAL(line-1,col-1) + SVAL(line-1,col) + SVAL(line-1,col+1);
           sum += SVAL(line,col-1) + SVAL(line,col+1);
           sum += SVAL(line+1,col-1) + SVAL(line+1,col) + SVAL(line+1,col+1);
           
           if (SVAL(line,col)) {
               //-------------------
               // die or survive ???
               //-------------------
               if (sum > 1 && sum < 4) DVAL(line,col) = 1;
               else                    DVAL(line,col) = 0;
           }
           else {
               //-------------------------------------------
               // birth of a new member of this universe ??
               // or still a waste cell ??
               //-------------------------------------------
               if (sum == 3) DVAL(line,col) = 1;
               else          DVAL(line,col) = 0;
           }                      
       }
   }

   //-------------------------------------------------
   // handle top line with wrap around to bottom line
   // AND
   // handle bottom line with wrap around to top line
   //-------------------------------------------------
   for (col=1;col<maxcol;col++) {
       int sumtop,sumbottom;
       
       sumtop = SVAL(maxline,col-1) + SVAL(maxline,col) + SVAL(maxline,col+1) +
                SVAL(0,col-1)       + SVAL(0,col+1)     +
                SVAL(1,col-1)       + SVAL(1,col)       + SVAL(1,col+1);

       sumbottom = SVAL(maxline-1,col-1) + SVAL(maxline-1,col) + SVAL(maxline-1,col+1) +
                   SVAL(maxline,col-1)   + SVAL(maxline,col+1) +
                   SVAL(0,col-1)         + SVAL(0,col)         + SVAL(0,col+1);

       if (SVAL(0,col)) {
           if (sumtop > 1 && sumtop < 4) DVAL(0,col) = 1;
           else                          DVAL(0,col) = 0;
       }
       else {
           if (sumtop == 3) DVAL(0,col) = 1;
           else             DVAL(0,col) = 0;
       }                      

       if (SVAL(maxline,col)) {
           if (sumbottom > 1 && sumbottom < 4) DVAL(maxline,col) = 1;
           else                                DVAL(maxline,col) = 0;
       }
       else {
           if (sumbottom == 3) DVAL(maxline,col) = 1;
           else                DVAL(maxline,col) = 0;
       }                      
   }
   
   //--------------------------------------------------------------
   // handle most-left column with wrap around to most-right column
   // AND
   // handle most-right column with wrap around to most-left column
   //--------------------------------------------------------------
   for (line=1;line<maxline;line++) {
       int sumleft,sumright;
       
       sumleft  = SVAL(line-1,maxcol) + SVAL(line-1,0) + SVAL(line-1,1) +
                  SVAL(line,  maxcol) + SVAL(line,1) +
                  SVAL(line+1,maxcol) + SVAL(line+1,0) + SVAL(line+1,1);

       sumright = SVAL(line-1,maxcol-1) + SVAL(line-1,maxcol) + SVAL(line-1,0) +
                  SVAL(line,  maxcol-1) + SVAL(line,0) +
                  SVAL(line+1,maxcol-1) + SVAL(line+1,maxcol) + SVAL(line+1,0);
           
       if (SVAL(line,0)) {
           if (sumleft > 1 && sumleft < 4) DVAL(line,0) = 1;
           else                            DVAL(line,0) = 0;
       }
       else {
           if (sumleft == 3) DVAL(line,0) = 1;
           else              DVAL(line,0) = 0;
       }                      

       if (SVAL(line,maxcol)) {
           if (sumright > 1 && sumright < 4) DVAL(line,maxcol) = 1;
           else                              DVAL(line,maxcol) = 0;
       }
       else {
           if (sumright == 3) DVAL(line,maxcol) = 1;
           else               DVAL(line,maxcol) = 0;
       }                      
   }
}


//=============================================================================
// build our special "one pixel per byte" array from original LCD content
//=============================================================================
void GrabScreen() {
    unsigned char* lcdmem = LCD_MEM;    
    unsigned char* tmp    = source;
    int i;
    
    for (i=0;i<LCD_SIZE;i++) {
        if (*lcdmem & 0x80) *tmp++ = 1;
        else                *tmp++ = 0;
        if (*lcdmem & 0x40) *tmp++ = 1;
        else                *tmp++ = 0;
        if (*lcdmem & 0x20) *tmp++ = 1;
        else                *tmp++ = 0;
        if (*lcdmem & 0x10) *tmp++ = 1;
        else                *tmp++ = 0;
        if (*lcdmem & 0x08) *tmp++ = 1;
        else                *tmp++ = 0;
        if (*lcdmem & 0x04) *tmp++ = 1;
        else                *tmp++ = 0;
        if (*lcdmem & 0x02) *tmp++ = 1;
        else                *tmp++ = 0;
        if (*lcdmem & 0x01) *tmp++ = 1;
        else                *tmp++ = 0;
        lcdmem++;
    }
}    


//=============================================================================
// write our special "one pixel per byte" array to LCD buffer
//=============================================================================
void PushScreen() {
    unsigned char* lcdmem = LCD_MEM;    
    unsigned char* tmp    = dest;
    int i;
    
    for (i=0;i<LCD_SIZE;i++) {
        *lcdmem = 0;
        if (*tmp++) *lcdmem |= 0x80;
        if (*tmp++) *lcdmem |= 0x40;
        if (*tmp++) *lcdmem |= 0x20;
        if (*tmp++) *lcdmem |= 0x10;
        if (*tmp++) *lcdmem |= 0x08;
        if (*tmp++) *lcdmem |= 0x04;
        if (*tmp++) *lcdmem |= 0x02;
        if (*tmp++) *lcdmem |= 0x01;
        lcdmem++;
    }
}    


//=============================================================================
// where all the fun start as usual ...
//=============================================================================
void _main(void) {
    unsigned int counter = 0;
    char         outbuf[100];
    LCD_BUFFER   savescreen;   // NOTE: I do not use the SAVESCREEN_PATCH, 
                               //       because I want to output a message to
                               //       the statusline at the end
    
    //------------------------------------------------------
    // allocate first "one pixel per byte" array buffer
    //------------------------------------------------------
    if (!(source = (unsigned char*)malloc(240*128))) {
        ST_showHelp("out of memory");
        return;
    }

    //------------------------------------------------------
    // allocate second "one pixel per byte" array buffer
    //------------------------------------------------------
    if (!(dest = (unsigned char*)malloc(240*128))) {
        ST_showHelp("out of memory");
        free(source);
        return;
    }
    
    LCD_save(savescreen);
    
    maxcol  = LCD_WIDTH  - 1;
    maxline = LCD_HEIGHT - 1;
    
    //------------------------------------------------------
    // fill up source array with data from screen
    //------------------------------------------------------
    GrabScreen();

    FontSetSys(F_4x6);


    //------------------------------------------------------
    // loop until user gets bored ...
    //------------------------------------------------------
    while (!kbhit()) {
       unsigned char* tmp;

       //-----------------------------------------------
       // do a full iteration, push data back on screen,
       // output the iteration counter and swap source
       // and dest pointers
       //-----------------------------------------------
       Process();
       PushScreen();
       counter++;
       sprintf(outbuf,"%05u",counter);
       DrawStrXY(0,0,outbuf,A_REPLACE);
       tmp = dest;
       dest = source;
       source = tmp;
    }
    
    //------------------------------------------------------
    // fetch key which is already in the keyboard buffer,
    // free allocated memory and get out here !!
    //------------------------------------------------------
    ngetchx();
    free(source);
    free(dest);
    LCD_restore(savescreen);
    ST_showHelp("TICT Production 2000");
}

//#############################################################################
// Revision History
//#############################################################################
//
// $Log: life.c,v $
// Revision 1.1  2000/08/28 17:50:34  Thomas Nussbaumer
// initial version
//
//
//
//
